;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copyright, 1990. The Regents of the University of California.
;;; This software was produced under a U. S. Government contract
;;; (W-7405-ENG-36) by Los Alamos National Laboratory, which is
;;; operated by the University of California for the U. S. Department
;;; of Energy.  The U. S. Government is licensed to use, reproduce,
;;; and distribute this software. Permission is granted to the
;;; public to copy and use this software without charge, provided
;;; that this Notice and any statement of authorship are reproduced
;;; on all copies.  Neither the Government nor the University makes
;;; any warranty, express or implied, or assumes any liability or
;;; responsibility for the use of this software.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'kee)

(goto.kb 'clos-on-kee)

;;;
;;; class-of. Instances created by make-instance will have only one
;;; parent.
;;;

(defun class-of (instance)
  "Get the class of an instance."
  (let ((unit (unit.exists.p instance)))
    (if (null unit)
	()
	(car (unit.parents unit 'member)))))

;;;
;;; find-class turns a symbol into a class object. KEE will get the unit for
;;; us. However, the only way to tell if the KEE unit is actually a class is to
;;; start at the class 'standard-object, and walk down the tree of all its
;;; subclasses. KEE gives us just what we need...
;;;

(defun find-class (symbol &optional (errorp t) environment)
  "Turn a symbol into its associated class object."
  (let ((unit (unit.exists.p symbol)))
    (if (null unit)
	(if errorp
	    (error "No class exists named ~a~%" symbol)
	    ())
	(let ((result (unit.descendant.p
			   unit
			   '(standard-object clos-on-kee)
			   'subclass)))
	  (if result
	      result
	      (if errorp
		  (error "No class exists named ~a~%" symbol)
		  errorp)))
	)))

;;;
;;; class-name and its setf method
;;;

(defun class-name (class)
  "Return the name of some class object."
  (unit.name class))

(defun class-name-setf-method (class new-name)
  "Change the name of a class."
  (rename.unit class new-name)
  new-name)

(defsetf class-name class-name-setf-method)

;;;
;;; slot-value and its setf method
;;;

(defun slot-value (instance slot-name)
  "The basic slot-accessor."
  (get.value instance slot-name))

(defun slot-value-setf-method (instance slot-name value)
  "A setf method for slot-value."
  (put.value instance slot-name value)
  value)

(defsetf slot-value slot-value-setf-method)

;;;
;;; Metaobject Protocol.
;;;
;;; class "slot" readers
;;;

(defun class-direct-subclasses (class)
  "return a list of all direct subclasses of this class."
  (unit.children class 'subclass))

(defun class-direct-superclasses (class)
  "return a list of all direct subclasses of this class."
  (unit.parents class 'subclass))

;;;
;;; Should we use the CLOS definition of class precedence list? or the
;;; KEE definition?
;;; I choose to use the KEE definition here so as to be consistant with
;;; the sort used by method combination. It should be similar enough to
;;; CLOS for all practical purposes.
;;;

(defun class-precedence-list (class)
  (cons class (butlast (unit.ancestors class 'subclass))))
